"""AggregatedFeatureSet entity."""
import itertools
from datetime import timedelta
from functools import reduce
from typing import Dict, List

from pyspark import sql
from pyspark.sql import DataFrame, functions

from butterfree.clients import SparkClient
from butterfree.dataframe_service import repartition_df
from butterfree.transform import FeatureSet
from butterfree.transform.features import Feature, KeyFeature, TimestampFeature
from butterfree.transform.transformations import AggregatedTransform
from butterfree.transform.utils import Window, date_range


class AggregatedFeatureSet(FeatureSet):
    """Holds metadata about the aggregated feature set.

    This class overrides some methods of the parent FeatureSet class and has specific
    methods for aggregations.

    The AggregatedTransform can only be used on AggregatedFeatureSets. The construct
    method will be responsible by collecting every feature's transformation definitions
    so it can run a groupby over the input dataframe, taking into account whether the
    user want's to run an rolling window aggregation, pivoting or just apply aggregation
    functions.

    Example:
        This an example regarding the aggregated feature set definition. All features
        and its transformations are defined.

    >>> from butterfree.transform.aggregated_feature_set import (
    ...       AggregatedFeatureSet
    ... )
    >>> from butterfree.transform.features import (
    ...     Feature,
    ...     KeyFeature,
    ...     TimestampFeature,
    ...)
    >>> from butterfree.transform.transformations import (
    ...     AggregatedTransform,
    ... )
    >>> from butterfree.constants import DataType
    >>> from butterfree.clients import SparkClient
    >>> from butterfree.transform.utils import Function
    >>> import pyspark.sql.functions as F
    >>> client = SparkClient()
    >>> client.conn.conf.set("spark.sql.session.timeZone", "UTC")
    >>> dataframe = client.conn.createDataFrame(
    ...     [
    ...         (1, "2020-01-01 13:01:00+000", 1000, "publicado"),
    ...         (2, "2020-01-01 14:01:00+000", 2000, "publicado"),
    ...         (1, "2020-01-02 13:01:00+000", 2000, "alugado"),
    ...         (1, "2020-01-03 13:01:00+000", 1000, "despublicado"),
    ...         (2, "2020-01-09 14:01:00+000", 1000, "despublicado"),
    ...     ],
    ...     ("id", "ts", "rent", "status"),
    ... )
    >>> dataframe = dataframe.withColumn("ts", dataframe["ts"].cast("timestamp"))
    >>> feature_set = AggregatedFeatureSet(
    ...    name="aggregated_feature_set",
    ...    entity="entity",
    ...    description="description",
    ...    features=[
    ...        Feature(
    ...            name="feature1",
    ...            description="test",
    ...            transformation=AggregatedTransform(
    ...                 functions=[
    ...                    Function(F.avg, DataType.DOUBLE),
    ...                    Function(F.stddev_pop, DataType.DOUBLE)],
    ...             ),
    ...             from_column="rent",
    ...        ),
    ...    ],
    ...    keys=[KeyFeature(name="id", description="lul")],
    ...    timestamp=TimestampFeature(from_column="ts"),
    ...)
    >>> result = feature_set.construct(
    ...     dataframe=dataframe,
    ...     client=client,
    ...     end_date="2020-01-15"
    ... )
    >>> result.show()
    +---+-------------------+-------------+----------------+
    | id|          timestamp|feature1__avg|feature1__stddev|
    +---+-------------------+-------------+----------------+
    |  1|2020-01-01 13:01:00|       1000.0|            null|
    |  1|2020-01-03 13:01:00|       1000.0|            null|
    |  1|2020-01-02 13:01:00|       2000.0|            null|
    |  2|2020-01-09 14:01:00|       1000.0|            null|
    |  2|2020-01-01 14:01:00|       2000.0|            null|
    +---+-------------------+-------------+----------------+

    Since you didn't define a window, the AggregateFeatureSet will always group by
    keys and timestamp feature columns. So in this example, there will be no changes to
    the dataframe, since it doesn't duplicate on id and timestamp :)

    Let's run one example with windows:

    >>> feature_set.with_windows(definitions=["3 days"])
    >>> result = feature_set.construct(
    ...     dataframe=dataframe,
    ...     client=client,
    ...     end_date="2020-01-15"
    ... )
    >>> result.orderBy("timestamp", "id").show()
    +---+-------------------+-----------------------------------------+
    | id|          timestamp|feature1__avg_over_3_days_rolling_windows|
    +---+-------------------+-----------------------------------------+
    |  1|2020-01-01 00:00:00|                                     null|
    |  2|2020-01-01 00:00:00|                                     null|
    |  1|2020-01-02 00:00:00|                                   1000.0|
    |  2|2020-01-02 00:00:00|                                   2000.0|
    |  1|2020-01-03 00:00:00|                                   1500.0|
    |  1|2020-01-04 00:00:00|                       1333.3333333333333|
    |  1|2020-01-05 00:00:00|                                   1500.0|
    |  2|2020-01-05 00:00:00|                                     null|
    |  1|2020-01-06 00:00:00|                                   1000.0|
    |  1|2020-01-07 00:00:00|                                     null|
    |  2|2020-01-10 00:00:00|                                   1000.0|
    |  2|2020-01-13 00:00:00|                                     null|
    +---+-------------------+-----------------------------------------+

    +---+-------------------+--------------------------------------------+
    | id|          timestamp|feature1__stddev_over_3_days_rolling_windows|
    +---+-------------------+--------------------------------------------+
    |  1|2020-01-01 00:00:00|                                        null|
    |  2|2020-01-01 00:00:00|                                        null|
    |  1|2020-01-02 00:00:00|                                        null|
    |  2|2020-01-02 00:00:00|                                        null|
    |  1|2020-01-03 00:00:00|                           707.1067811865476|
    |  1|2020-01-04 00:00:00|                           577.3502691896258|
    |  1|2020-01-05 00:00:00|                           707.1067811865476|
    |  2|2020-01-05 00:00:00|                                        null|
    |  1|2020-01-06 00:00:00|                                        null|
    |  1|2020-01-07 00:00:00|                                        null|
    |  2|2020-01-10 00:00:00|                                        null|
    |  2|2020-01-13 00:00:00|                                        null|
    +---+-------------------+--------------------------------------------+

    (Had to break down the table result.)

    And with pivot:

    >>> feature_set.with_pivot(column="status", values=["publicado", "despublicado"])
    +---+-------------------+-----------------------+--------------------------+
    | id|          timestamp|publicado_feature1__avg|publicado_feature1__stddev|
    +---+-------------------+-----------------------+--------------------------+
    |  1|2020-01-01 13:01:00|                 1000.0|                      null|
    |  2|2020-01-01 14:01:00|                 2000.0|                      null|
    |  1|2020-01-02 13:01:00|                   null|                      null|
    |  1|2020-01-03 13:01:00|                   null|                      null|
    |  2|2020-01-09 14:01:00|                   null|                      null|
    +---+-------------------+-----------------------+--------------------------+

    +---+-------------------+--------------------------+-----------------------------+
    | id|          timestamp|despublicado_feature1__avg|despublicado_feature1__stddev|
    +---+-------------------+--------------------------+-----------------------------+
    |  1|2020-01-01 13:01:00|                      null|                         null|
    |  2|2020-01-01 14:01:00|                      null|                         null|
    |  1|2020-01-02 13:01:00|                      null|                         null|
    |  1|2020-01-03 13:01:00|                    1000.0|                         null|
    |  2|2020-01-09 14:01:00|                    1000.0|                         null|
    +---+-------------------+--------------------------+-----------------------------+

    As you can see, we need to pass the values you want to pivot. It optimizes this
    processing in spark and allows ignoring values when pivoting. If you wanted to get
    the pivot aggregation for "alugado" too, just use:

    >>> feature_set.with_pivot(
    ...     column="status", values=["publicado", "despublicado", "alugado"]
    ... )

    You can also run it with pivot AND windows:

    >>> feature_set.with_pivot(
    ...     column="status",
    ...     values=["publicado", "despublicado", "alugado"]
    ... ).with_windows(definitions=["1 day", "2 weeks"])

    The construct method will execute the feature set, computing all the
    defined aggregated transformations at once.

    Remember: when using an AggregatedFeatureSet without window, the group will use the
    timestamp column.
    """

    def __init__(
        self,
        name: str,
        entity: str,
        description: str,
        keys: List[KeyFeature],
        timestamp: TimestampFeature,
        features: List[Feature],
    ):
        self._windows = []
        self._pivot_column = None
        self._pivot_values = []
        self._distinct_subset = []
        self._distinct_keep = None
        super(AggregatedFeatureSet, self).__init__(
            name, entity, description, keys, timestamp, features,
        )

    @property
    def features(self) -> List[Feature]:
        """Features to compose the feature set."""
        return self.__features

    @features.setter
    def features(self, value: List[Feature]):
        if not isinstance(value, list) or not all(
            isinstance(item, Feature) for item in value
        ):
            raise ValueError("features needs to be a list of Feature objects.")

        feature_columns = self._get_features_columns(*value)
        if len(feature_columns) != len(set(feature_columns)):
            raise KeyError("feature columns will have duplicates.")

        if not self._has_aggregated_transform_only(value):
            raise ValueError(
                "You can't define a aggregated feature without aggregated transform. "
                "You need to use Feature Set."
            )

        self.__features = value

    @staticmethod
    def _has_aggregated_transform_only(features):
        """Aggregated Transform check.

        Checks if all transformations are AggregatedTransform within the scope of the
        AggregatedFeatureSet.

        Returns:
            True if there's a aggregation transform.

        """
        return all(
            [
                isinstance(feature.transformation, AggregatedTransform)
                for feature in features
            ]
        )

    @staticmethod
    def _build_feature_column_name(feature_column, pivot_value=None, window=None):
        base_name = feature_column
        if pivot_value is not None:
            base_name = f"{pivot_value}_{base_name}"
        if window is not None:
            base_name = f"{base_name}_{window.get_name()}"

        return base_name

    @property
    def features_columns(self) -> List[str]:
        """Name of the columns of all features in feature set."""
        base_columns = list(
            itertools.chain(*[f.get_output_columns() for f in self.features])
        )

        pivot_values = self._pivot_values or [None]
        windows = self._windows or [None]
        feature_columns = [
            self._build_feature_column_name(fc, pivot_value=pv, window=w)
            for pv, fc, w in itertools.product(pivot_values, base_columns, windows)
        ]
        return feature_columns

    def with_distinct(self, subset: List, keep: str = "last"):
        """Add a distinct configuration for your aggregated feature set.

        Args:
            subset: the columns where it will identify duplicates.
            keep: determines which duplicates to keep. Default 'last'.
                - first : Ascending sorting by timestamp.
                - last : Descending sorting by timestamp.

        Returns:
            An AggregatedFeatureSet configured with distinct.
        """
        if keep not in ["last", "first"]:
            raise ValueError("The distinct keep param can only be 'last' or 'first'.")

        if not subset:
            raise ValueError("The distinct subset param can't be empty.")

        self._distinct_subset = subset
        self._distinct_keep = keep

        return self

    def with_windows(self, definitions: List[str]):
        """Create a list with windows defined."""
        self._windows = [
            Window(
                partition_by=None,
                order_by=None,
                mode="rolling_windows",
                window_definition=definition,
            )
            for definition in definitions
        ]
        return self

    def with_pivot(self, column: str, values: List[str]):
        """Add a pivot configuration for your aggregated feature set.

        This means we will group the input data, pivot over the column parameter and
        run each aggregation function over the columns within each value group in
        the values parameter. In spark it will be something like:
        dataframe.groupBy(*group).pivot(column, values).agg(*aggregations)

        Args:
            column: the column, containing categorical values, to pivot on.
            values: the distinct values you want to be pivoted.

        Returns:
            An AggregatedFeatureSet configured with pivot.
        """
        self._pivot_column = column
        self._pivot_values = values
        return self

    def _get_base_dataframe(self, client, dataframe, end_date):
        start_date = dataframe.agg(functions.min(self.timestamp_column)).take(1)[0][0]
        end_date = end_date or dataframe.agg(functions.max(self.timestamp_column)).take(
            1
        )[0][0] + timedelta(days=1)
        date_df = date_range.get_date_range(client, start_date, end_date)
        unique_keys = dataframe.dropDuplicates(subset=self.keys_columns).select(
            *self.keys_columns
        )

        return unique_keys.crossJoin(date_df)

    @staticmethod
    def _dataframe_join(left, right, on, how, num_processors=None):
        # make both tables co-partitioned to improve join performance
        left = repartition_df(left, partition_by=on, num_processors=num_processors)
        right = repartition_df(right, partition_by=on, num_processors=num_processors)
        return left.join(right, on=on, how=how)

    def _aggregate(self, dataframe, features, window=None, num_processors=None):
        aggregations = [
            c.function for f in features for c in f.transformation.aggregations
        ]

        groupby = self.keys_columns.copy()
        if window is not None:
            dataframe = dataframe.withColumn("window", window.get())
            groupby.append("window")
        else:
            groupby.append(self.timestamp_column)

        if self._distinct_subset:
            orderby = (
                functions.col(self.timestamp_column).desc()
                if self._distinct_keep == "last"
                else functions.col(self.timestamp_column).asc()
            )

            partition_window = (
                sql.Window()
                .partitionBy(*groupby, *self._distinct_subset)
                .orderBy(orderby)
            )

            dataframe = dataframe.withColumn(
                "keep_rn", functions.row_number().over(partition_window)
            ).filter("keep_rn = 1")

        # repartition to have all rows for each group at the same partition
        # by doing that, we won't have to shuffle data on grouping by id
        dataframe = repartition_df(
            dataframe, partition_by=groupby, num_processors=num_processors,
        )
        grouped_data = dataframe.groupby(*groupby)

        if self._pivot_column:
            grouped_data = grouped_data.pivot(self._pivot_column, self._pivot_values)

        aggregated = grouped_data.agg(*aggregations)
        return self._with_renamed_columns(aggregated, features, window)

    def _with_renamed_columns(self, aggregated, features, window):
        old_columns = [
            c
            for c in aggregated.columns
            if c not in self.keys_columns + [self.timestamp_column, "window"]
        ]

        # the loop order here is really important
        # spark will always create the aggregated dataframe following the order:
        # pivot_value then aggregation function
        # so when we have 2 pivot values and 2 agg functions
        # the result columns will be something like: pivot_value1_function1,
        # pivot_value1_function_2, pivot_value2_function1 and pivot_value2_function2
        base_columns = list(
            itertools.chain(*[f.get_output_columns() for f in features])
        )
        pivot_values = self._pivot_values or [None]
        new_columns = [
            self._build_feature_column_name(fc, pivot_value=pv, window=window)
            for pv, fc in itertools.product(pivot_values, base_columns)
        ]

        type = [c.data_type for f in features for c in f.transformation.aggregations]
        if pivot_values != [None]:
            type = len(pivot_values) * type

        select = [
            f"cast(`{old_name}` as {dt.simpleString()}) as {new_name}"
            for old_name, new_name, dt in zip(old_columns, new_columns, type)
        ]
        if self._windows:
            select.append(f"window.end as {self.timestamp_column}")
        else:
            select.append(f"{self.timestamp_column}")

        select += [kc for kc in self.keys_columns]
        return aggregated.selectExpr(*select)

    def get_schema(self) -> List[Dict]:
        """Get feature set schema.

        Returns:
            List of dicts with the feature set schema.

        """
        schema = []
        for f in self.keys + [self.timestamp]:
            for c in self._get_features_columns(f):
                schema.append(
                    {
                        "column_name": c,
                        "type": f.dtype.spark,
                        "primary_key": True if isinstance(f, KeyFeature) else False,
                    }
                )
        pivot_values = self._pivot_values or [None]
        windows = self._windows or [None]

        for f in self.features:
            combination = itertools.product(
                pivot_values, self._get_features_columns(f), windows
            )
            name = [
                self._build_feature_column_name(fc, pivot_value=pv, window=w)
                for pv, fc, w in combination
            ]
            type = [
                fc.data_type.spark
                for fc in f.transformation.functions
                for _ in range(len(pivot_values) * len(windows))
            ]

            for n, dt in zip(name, type):
                schema.append({"column_name": n, "type": dt, "primary_key": False})

        return schema

    def construct(
        self,
        dataframe: DataFrame,
        client: SparkClient,
        end_date: str = None,
        num_processors: int = None,
    ) -> DataFrame:
        """Use all the features to build the feature set dataframe.

        After that, there's the caching of the dataframe, however since cache()
        in Spark is lazy, an action is triggered in order to force persistence,
        but we only cache if it is not a streaming spark dataframe.

        Args:
            dataframe: input dataframe to be transformed by the features.
            client: client responsible for connecting to Spark session.
            end_date: user defined max date for having aggregated data (exclusive).
            num_processors: cluster total number of processors for repartitioning.

        Returns:
            Spark dataframe with all the feature columns.

        """
        if end_date is None and self._windows:
            raise ValueError(
                "When using aggregate with windows, one must give end_date."
            )

        if not isinstance(dataframe, DataFrame):
            raise ValueError("source_df must be a dataframe")

        output_df = reduce(
            lambda df, feature: feature.transform(df),
            self.keys + [self.timestamp],
            dataframe,
        )

        if self._windows:
            # prepare our left table, a cartesian product between distinct keys
            # and dates in range for this feature set
            base_df = self._get_base_dataframe(
                client=client, dataframe=output_df, end_date=end_date
            )

            # run aggregations for each window
            agg_list = [
                self._aggregate(
                    dataframe=output_df,
                    features=self.features,
                    window=w,
                    num_processors=num_processors,
                )
                for w in self._windows
            ]

            # left join each aggregation result to our base dataframe
            output_df = reduce(
                lambda left, right: self._dataframe_join(
                    left,
                    right,
                    on=self.keys_columns + [self.timestamp_column],
                    how="left",
                    num_processors=num_processors,
                ),
                agg_list,
                base_df,
            )
        else:
            output_df = self._aggregate(output_df, features=self.features)

        output_df = output_df.select(*self.columns).replace(float("nan"), None)
        if not output_df.isStreaming:
            output_df = self._filter_duplicated_rows(output_df)
            output_df.cache().count()

        return output_df



"""Abstract Validation class."""
from abc import ABC, abstractmethod

from pyspark.sql.dataframe import DataFrame


class Validation(ABC):
    """Validate dataframe properties.

    Attributes:
        dataframe: data to be verified.

    """

    def __init__(self, dataframe: DataFrame = None):
        self.dataframe = dataframe

    def input(self, dataframe: DataFrame):
        """Input a dataframe to check.

        Args:
            dataframe: data to check.

        """
        self.dataframe = dataframe
        return self

    @abstractmethod
    def check(self) -> None:
        """Check validation properties about the dataframe.

        Raises:
            ValueError: if any of the verifications fail.

        """



"""Writer entity."""

from abc import ABC, abstractmethod
from functools import reduce
from typing import Callable

from pyspark.sql.dataframe import DataFrame

from butterfree.clients import SparkClient
from butterfree.transform import FeatureSet


class Writer(ABC):
    """Abstract base class for Writers.

    Args:
        spark_client: client for spark connections with external services.

    """

    def __init__(self):
        self.transformations = []

    def with_(self, transformer: Callable, *args, **kwargs):
        """Define a new transformation for the Writer.

        All the transformations are used when the method consume is called.

        Args:
            transformer: method that receives a dataframe and output a
                dataframe.
            *args: args for the transformer.
            **kwargs: kwargs for the transformer.

        Returns:
            Reader object with new transformation

        """
        new_transformation = {
            "transformer": transformer,
            "args": args if args else (),
            "kwargs": kwargs if kwargs else {},
        }
        self.transformations.append(new_transformation)
        return self

    def _apply_transformations(self, df: DataFrame) -> DataFrame:
        return reduce(
            lambda result_df, transformation: transformation["transformer"](
                result_df, *transformation["args"], **transformation["kwargs"]
            ),
            self.transformations,
            df,
        )

    @abstractmethod
    def write(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient,
    ):
        """Loads the data from a feature set into the Feature Store.

        Feature Store could be Online or Historical.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        """

    @abstractmethod
    def validate(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient
    ):
        """Calculate dataframe rows to validate data into Feature Store.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        Raises:
            AssertionError: if validation fails.

        """



"""Holds the Online Feature Store writer class."""

import os
from typing import Any, List, Optional

from pyspark.sql import DataFrame, Window
from pyspark.sql.functions import col, row_number
from pyspark.sql.streaming import StreamingQuery

from butterfree.clients import SparkClient
from butterfree.configs.db import CassandraConfig
from butterfree.constants.columns import TIMESTAMP_COLUMN
from butterfree.load.writers.writer import Writer
from butterfree.transform import FeatureSet


class OnlineFeatureStoreWriter(Writer):
    """Enable writing feature sets into the Online Feature Store.

    Attributes:
        db_config: Spark configuration for connect databases.
            For more information check the module 'butterfree.db.configs'.
        debug_mode: "dry run" mode, write the result to a temporary view.
        write_to_entity: option to write the data to the entity table.
            With this option set to True, the writer will write the feature set to
            a table with the name equal to the entity name, defined on the pipeline.
            So, it WILL NOT write to a table with the name of the feature set, as it
            normally does.

    Example:
        Simple example regarding OnlineFeatureStoreWriter class instantiation.
        We can instantiate this class without db configurations, so the class get the
        CassandraConfig() where it provides default configurations about CassandraDB.

    >>> spark_client = SparkClient()
    >>> writer = OnlineFeatureStoreWriter()
    >>> writer.write(feature_set=feature_set,
       ...           dataframe=dataframe,
       ...           spark_client=spark_client)

        However, we can define the db configurations and provide them to
        OnlineFeatureStoreWriter.

    >>> spark_client = SparkClient()
    >>> config = CassandraConfig(mode="overwrite",
        ...                      format_="parquet",
        ...                      keyspace="keyspace_name")

    >>> writer = OnlineFeatureStoreWriter(db_config=config)
    >>> writer.write(feature_set=feature_set,
       ...           dataframe=dataframe,
       ...           spark_client=spark_client)

        For what settings you can use on CassandraConfig and default settings,
        to read CassandraConfig class.

        We can instantiate OnlineFeatureStoreWriter class to validate the writers,
        using the default or custom configs.

    >>> spark_client = SparkClient()
    >>> writer = OnlineFeatureStoreWriter()
    >>> writer.validate(feature_set=feature_set,
       ...              dataframe=dataframe,
       ...              spark_client=spark_client)

        Both methods (writer and validate) will need the Spark Client,
        Feature Set and DataFrame, to write or to validate,
        according to OnlineFeatureStoreWriter class arguments.
    """

    __name__ = "Online Feature Store Writer"

    def __init__(self, db_config=None, debug_mode: bool = False, write_to_entity=False):
        super().__init__()
        self.db_config = db_config or CassandraConfig()
        self.debug_mode = debug_mode
        self.write_to_entity = write_to_entity

    @staticmethod
    def filter_latest(dataframe: DataFrame, id_columns: List[Any]) -> DataFrame:
        """Filters latest data from the dataframe.

        Args:
            dataframe: spark dataframe containing data from a feature set.
            id_columns: unique identifier column set for this feature set.

        Returns:
            dataframe: contains only latest data for each unique id in the
                feature set.

        """
        if TIMESTAMP_COLUMN not in dataframe.columns:
            raise KeyError("DataFrame must have a 'ts' column to order by.")
        if id_columns is None or not id_columns:
            raise ValueError("Users must provide the unique identifiers.")
        missing = [c for c in id_columns if c not in dataframe.columns]
        if missing:
            raise KeyError(f"{missing} not found in the DataFrame.")

        window = Window.partitionBy(*id_columns).orderBy(col(TIMESTAMP_COLUMN).desc())
        return (
            dataframe.select(col("*"), row_number().over(window).alias("rn"),)
            .filter(col("rn") == 1)
            .drop("rn")
        )

    def _write_stream(
        self,
        feature_set: FeatureSet,
        dataframe: DataFrame,
        spark_client: SparkClient,
        table_name: str,
    ) -> StreamingQuery:
        """Writes the dataframe in streaming mode."""
        checkpoint_folder = (
            f"{feature_set.name}__on_entity" if self.write_to_entity else table_name
        )
        checkpoint_path = (
            os.path.join(
                self.db_config.stream_checkpoint_path,
                feature_set.entity,
                checkpoint_folder,
            )
            if self.db_config.stream_checkpoint_path
            else None
        )
        streaming_handler = spark_client.write_stream(
            dataframe,
            processing_time=self.db_config.stream_processing_time,
            output_mode=self.db_config.stream_output_mode,
            checkpoint_path=checkpoint_path,
            format_=self.db_config.format_,
            mode=self.db_config.mode,
            **self.db_config.get_options(table_name),
        )
        return streaming_handler

    @staticmethod
    def _write_in_debug_mode(
        table_name: str, dataframe: DataFrame, spark_client: SparkClient
    ) -> Optional[StreamingQuery]:
        """Creates a temporary table instead of writing to the real feature store."""
        return spark_client.create_temporary_view(
            dataframe=dataframe, name=f"online_feature_store__{table_name}"
        )

    def write(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient,
    ) -> Optional[StreamingQuery]:
        """Loads the latest data from a feature set into the Feature Store.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        Returns:
            Streaming handler if writing streaming df, None otherwise.

        If the debug_mode is set to True, a temporary table with a name in the format:
        `online_feature_store__my_feature_set` will be created instead of writing to
        the real online feature store. If dataframe is streaming this temporary table
        will be updated in real time.

        """
        table_name = feature_set.entity if self.write_to_entity else feature_set.name

        if dataframe.isStreaming:
            dataframe = self._apply_transformations(dataframe)
            if self.debug_mode:
                return self._write_in_debug_mode(
                    table_name=table_name,
                    dataframe=dataframe,
                    spark_client=spark_client,
                )
            return self._write_stream(
                feature_set=feature_set,
                dataframe=dataframe,
                spark_client=spark_client,
                table_name=table_name,
            )

        latest_df = self.filter_latest(
            dataframe=dataframe, id_columns=feature_set.keys_columns
        )

        latest_df = self._apply_transformations(latest_df)

        if self.debug_mode:
            return self._write_in_debug_mode(
                table_name=table_name, dataframe=latest_df, spark_client=spark_client
            )

        return spark_client.write_dataframe(
            dataframe=latest_df,
            format_=self.db_config.format_,
            mode=self.db_config.mode,
            **self.db_config.get_options(table_name),
        )

    def validate(self, feature_set: FeatureSet, dataframe, spark_client: SparkClient):
        """Calculate dataframe rows to validate data into Feature Store.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        Raises:
            AssertionError: if validation fails.

        """
        # validation with online feature store can be done right now, since
        # this database can be updated by an streaming job while our ETL is running
        # therefore, creating in consistencies between this ETL feature set count and
        # the data already written to the FS.
        # TODO how to run data validations when a database has concurrent writes.
        pass

    def get_db_schema(self, feature_set: FeatureSet):
        """Get desired database schema.

        Args:
            feature_set: object processed with feature set metadata.

        Returns:
            Desired database schema.

        """
        db_schema = self.db_config.translate(feature_set.get_schema())
        return db_schema



"""Method to compute most frequent set aggregation."""
import pandas as pd
from pyspark.sql.functions import pandas_udf
from pyspark.sql.types import ArrayType, StringType


@pandas_udf(ArrayType(StringType()))
def most_frequent_set(column: pd.Series) -> list:
    """Computes the most frequent set aggregation.

    Attributes:
        column: desired data to be aggregated with most frequent set aggregation.

    Example:
        It's necessary to declare the desired aggregation method, (average,
        standard deviation and count are currently supported, as it can be
        seen in __ALLOWED_AGGREGATIONS) and define the most frequent set aggregation.

        >>> from pyspark import SparkContext
        >>> from pyspark.sql import session, Window
        >>> from butterfree.transform\
        ...     .transformations.user_defined_functions import (most_frequent_set)
        >>> sc = SparkContext.getOrCreate()
        >>> spark = session.SparkSession(sc)
        >>> df = spark.createDataFrame(
        >>>      [(1, 1), (1, 1), (2, 2), (2, 1), (2, 2)],
        >>>      ("id", "column"))
        >>> df.groupby("id").agg(most_frequent_set("column")).show()
        +---+-------------------------+
        | id|most_frequent_set(column)|
        +---+-------------------------+
        |  1|                      [1]|
        |  2|                   [2, 1]|
        +---+-------------------------+
        >>> w = Window.partitionBy('id').rowsBetween(
        ...        Window.unboundedPreceding, Window.unboundedFollowing)
        >>> df.withColumn(
        ...     'most_viewed', most_frequent_set("column").over(w)
        ... ).show()
        +---+------+-----------+
        | id|column|most_viewed|
        +---+------+-----------+
        |  1|     1|        [1]|
        |  1|     1|        [1]|
        |  2|     2|     [2, 1]|
        |  2|     1|     [2, 1]|
        |  2|     2|     [2, 1]|
        +---+------+-----------+

        This example shows the mode aggregation. It returns a list with the most
        frequent values. It's important to notice, however, that if we want to
        use it in fixed_windows or row_windows mode, we'd need unbounded windows.
        For that reason, mode is meant to be used just in rolling_windows mode,
        initially. We intend to make it available to others modes soon.

    """
    return column.astype(str).value_counts().index.tolist()



"""Reader entity."""

from abc import ABC, abstractmethod
from functools import reduce
from typing import Callable, List

from pyspark.sql import DataFrame

from butterfree.clients import SparkClient


class Reader(ABC):
    """Abstract base class for Readers.

    Attributes:
        id: unique string id for register the reader as a view on the metastore.
        transformations: list os methods that will be applied over the dataframe
            after the raw data is extracted.

    """

    def __init__(self, id: str):
        self.id = id
        self.transformations = []

    def with_(self, transformer: Callable, *args, **kwargs):
        """Define a new transformation for the Reader.

        All the transformations are used when the method consume is called.

        Args:
            transformer: method that receives a dataframe and output a
                dataframe.
            *args: args for the transformer.
            **kwargs: kwargs for the transformer.

        Returns:
            Reader object with new transformation

        """
        new_transformation = {
            "transformer": transformer,
            "args": args if args else (),
            "kwargs": kwargs if kwargs else {},
        }
        self.transformations.append(new_transformation)
        return self

    def _apply_transformations(self, df: DataFrame) -> DataFrame:
        return reduce(
            lambda result_df, transformation: transformation["transformer"](
                result_df, *transformation["args"], **transformation["kwargs"]
            ),
            self.transformations,
            df,
        )

    @abstractmethod
    def consume(self, client: SparkClient) -> DataFrame:
        """Extract data from target origin.

        Args:
            client: client responsible for connecting to Spark session.

        Returns:
            Dataframe with all the data.

        :return: Spark dataframe
        """

    def build(self, client: SparkClient, columns: List[tuple] = None):
        """Register the data got from the reader in the Spark metastore.

        Create a temporary view in Spark metastore referencing the data
        extracted from the target origin after the application of all the
        defined pre-processing transformations.

        Args:
            client: client responsible for connecting to Spark session.
            columns: list of tuples for renaming/filtering the dataset.

        """
        transformed_df = self._apply_transformations(self.consume(client))

        if columns:
            select_expression = []
            for old_expression, new_column_name in columns:
                select_expression.append(f"{old_expression} as {new_column_name}")
            transformed_df = transformed_df.selectExpr(*select_expression)

        transformed_df.createOrReplaceTempView(self.id)



"""TableSource entity."""

from pyspark.sql import DataFrame

from butterfree.clients import SparkClient
from butterfree.extract.readers.reader import Reader


class TableReader(Reader):
    """Responsible for get data from tables registered in the metastore.

    Attributes:
        id: unique string id for register the reader as a view on the metastore.
        database: name of the metastore database/schema.
        table: name of the table.

    Example:
        Simple example regarding TableReader class instantiation.

    >>> from butterfree.extract.readers import TableReader
    >>> from butterfree.clients import SparkClient
    >>> from butterfree.extract.pre_processing import filter
    >>> spark_client = SparkClient()
    >>> table_reader = TableReader(
    ...                     id="table_reader_id",
    ...                     database="table_reader_db",
    ...                     table="table_reader_table"
    ...                )
    >>> df = table_reader.consume(spark_client)

        This last method will use the Spark Client, as default, to read
        the desired table, loading data into a dataframe, according to
        TableReader class arguments.

        It's also possible to define simple transformations within the
        reader's scope:

    >>> table_reader.with_(filter, condition="year = 2019").build(spark_client)

        In this case, however, a temp view will be created, cointaining
        the transformed data.

    """

    __name__ = "Table Reader"

    def __init__(self, id: str, table: str, database: str = None):
        super().__init__(id)
        if not isinstance(table, str):
            raise ValueError(
                "table needs to be a string with the name of the registered table"
            )
        self.database = database
        self.table = table

    def consume(self, client: SparkClient) -> DataFrame:
        """Extract data from a table in Spark metastore.

        Args:
            client: client responsible for connecting to Spark session.

        Returns:
            Dataframe with all the data from the table.

        """
        return client.read_table(self.table, self.database)



"""This module holds database configurations to be used by clients."""

from butterfree.configs.db.abstract_config import AbstractWriteConfig
from butterfree.configs.db.cassandra_config import CassandraConfig
from butterfree.configs.db.kafka_config import KafkaConfig
from butterfree.configs.db.metastore_config import MetastoreConfig

__all__ = ["AbstractWriteConfig", "CassandraConfig", "KafkaConfig", "MetastoreConfig"]



"""Holds connection clients."""
from butterfree.clients.abstract_client import AbstractClient
from butterfree.clients.cassandra_client import CassandraClient
from butterfree.clients.spark_client import SparkClient

__all__ = ["SparkClient", "CassandraClient", "AbstractClient"]



"""Module where filter DataFrames coming from readers."""
from pyspark.sql.dataframe import DataFrame


def filter(dataframe: DataFrame, condition):
    """Filters DataFrame's rows using the given condition and value.

    Args:
        dataframe: Spark DataFrame.
        condition: SQL expression with column, operation and value
            to filter the dataframe.

    Returns:
        Filtered dataframe
    """
    if not isinstance(condition, str):
        raise TypeError("condition should be string.")

    return dataframe.filter(condition)



"""Abstract classes for database configurations with spark."""

from abc import ABC, abstractmethod
from typing import Dict, List


class AbstractWriteConfig(ABC):
    """Abstract class for database write configurations with spark."""

    @property
    @abstractmethod
    def mode(self) -> str:
        """Config option "mode" for spark write.

        Args:

        Returns:
          str: mode.

        """

    @property
    @abstractmethod
    def format_(self) -> str:
        """Config option "format" for spark write.

        Args:

        Returns:
          str: format.

        """

    @abstractmethod
    def get_options(self, *args, **kwargs) -> dict:
        """Get connection options configuration defined in the entity.

        Args:
          *args: args to use in the options
          **kwargs: kwargs to use in the options

        Returns:
            Connection options configuration.

        """

    @abstractmethod
    def translate(self, schema) -> List[Dict]:
        """Translate feature set spark schema to the corresponding database.

        Args:
          schema: feature set schema

        Returns:
            Corresponding database schema.

        """



"""Holds configurations to read and write with Spark to AWS S3."""

import os
from typing import Dict, List

from butterfree.configs import environment
from butterfree.configs.db import AbstractWriteConfig


class MetastoreConfig(AbstractWriteConfig):
    """Configuration for Spark metastore database stored.

    By default the configuration is for AWS S3.

    Attributes:
        path: database root location.
        mode: writing mode used be writers.
        format_: expected stored file format.
        file_system: file schema uri, like: s3a, file.

    """

    def __init__(
        self,
        path: str = None,
        mode: str = None,
        format_: str = None,
        file_system: str = None,
    ):
        self.path = path
        self.mode = mode
        self.format_ = format_
        self.file_system = file_system

    @property
    def path(self) -> str:
        """Bucket name."""
        return self.__path

    @path.setter
    def path(self, value: str):
        self.__path = value or environment.get_variable("FEATURE_STORE_S3_BUCKET")

    @property
    def format_(self) -> str:
        """Expected stored file format."""
        return self.__format

    @format_.setter
    def format_(self, value: str):
        self.__format = value or "parquet"

    @property
    def mode(self) -> str:
        """Writing mode used be writers."""
        return self.__mode

    @mode.setter
    def mode(self, value):
        self.__mode = value or "overwrite"

    @property
    def file_system(self) -> str:
        """Writing mode used be writers."""
        return self.__file_system

    @file_system.setter
    def file_system(self, value):
        self.__file_system = value or "s3a"

    def get_options(self, key: str) -> dict:
        """Get options for Metastore.

        Options will be a dictionary with the write and read configuration for
        Spark Metastore.

        Args:
            key: path to save data into Metastore.

        Returns:
            Options configuration for Metastore.

        """
        return {
            "mode": self.mode,
            "format_": self.format_,
            "path": os.path.join(f"{self.file_system}://{self.path}/", key),
        }

    def translate(self, schema) -> List[Dict]:
        """Translate feature set spark schema to the corresponding database."""
        pass



"""Holds all transformations to be used by Features.

A transformation must inherit from a TransformComponent and handle data modification,
renaming and cast types using parent's (a Feature) information.
"""

from butterfree.transform.transformations.user_defined_functions.mode import mode
from butterfree.transform.transformations.user_defined_functions.most_frequent_set import (  # noqa
    most_frequent_set,
)

__all__ = [
    "mode",
    "most_frequent_set",
]



"""Methods to assert properties regarding Apache Spark Dataframes."""
from json import dumps
from typing import List

from pyspark import SparkContext
from pyspark.sql import Column, DataFrame, SparkSession
from pyspark.sql.functions import col


def assert_dataframe_equality(output_df: DataFrame, target_df: DataFrame):
    """Dataframe comparison method."""
    if not (
        output_df.count() == target_df.count()
        and len(target_df.columns) == len(output_df.columns)
    ):
        raise AssertionError(
            f"DataFrame shape mismatch: \n"
            f"output_df shape: {len(output_df.columns)} columns and "
            f"{output_df.count()} rows\n"
            f"target_df shape: {len(target_df.columns)} columns and "
            f"{target_df.count()} rows."
        )

    select_cols = [col(c) for c in output_df.schema.fieldNames()]

    output_data = sorted(output_df.select(*select_cols).collect())
    output_data = [row.asDict(recursive=True) for row in output_data]

    target_data = sorted(target_df.select(*select_cols).collect())
    target_data = [row.asDict(recursive=True) for row in target_data]

    if not output_data == target_data:
        raise AssertionError(
            "DataFrames have different values:\n"
            f"output_df records: {output_data}\n"
            f"target_data records: {target_data}."
        )


def assert_column_equality(
    output_df: DataFrame,
    target_df: DataFrame,
    output_column: Column,
    target_column: Column,
):
    """Columns comparison method."""
    if not (
        output_df.select(output_column).count()
        == target_df.select(target_column).count()
        and len(target_df.columns) == len(output_df.columns)
    ):
        raise AssertionError(
            f"DataFrame shape mismatch: \n"
            f"output_df shape: {len(output_df.columns)} columns and "
            f"{output_df.count()} rows\n"
            f"target_df shape: {len(target_df.columns)} columns and "
            f"{target_df.count()} rows."
        )

    output_data = output_df.selectExpr(f"{output_column} as {target_column}").collect()
    target_data = target_df.select(target_column).collect()
    if not output_data == target_data:
        raise AssertionError(
            "Columns have different values:\n"
            f"output_column records: {output_data}\n"
            f"target_column records: {target_data}."
        )


def create_df_from_collection(
    data: List[dict],
    spark_context: SparkContext,
    spark_session: SparkSession,
    schema=None,
):
    """Creates a dataframe from a list of dicts."""
    return spark_session.read.json(
        spark_context.parallelize(data, 1).map(lambda x: dumps(x)), schema=schema
    )



"""Holds configurations to read and write with Spark to Kafka."""
from typing import Dict, List

from butterfree.configs import environment
from butterfree.configs.db import AbstractWriteConfig


class KafkaConfig(AbstractWriteConfig):
    """Configuration for Spark to connect to Kafka.

    Attributes:
        kafka_topic: string with kafka topic name.
        kafka_connection_string: string with hosts and ports to connect.
        mode: write mode for Spark.
        format_: write format for Spark.
        stream_processing_time: processing time interval for streaming jobs.
        stream_output_mode: specify the mode from writing streaming data.
        stream_checkpoint_path: path on S3 to save checkpoints for the stream job.

    More information about processing_time, output_mode and checkpoint_path
    can be found in Spark documentation:
    [here](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html)

    """

    def __init__(
        self,
        kafka_topic: str = None,
        kafka_connection_string: str = None,
        mode: str = None,
        format_: str = None,
        stream_processing_time: str = None,
        stream_output_mode: str = None,
        stream_checkpoint_path: str = None,
    ):
        self.kafka_topic = kafka_topic
        self.kafka_connection_string = kafka_connection_string
        self.mode = mode
        self.format_ = format_
        self.stream_processing_time = stream_processing_time
        self.stream_output_mode = stream_output_mode
        self.stream_checkpoint_path = stream_checkpoint_path

    @property
    def kafka_topic(self) -> str:
        """Kafka topic name."""
        return self.__kafka_topic

    @kafka_topic.setter
    def kafka_topic(self, value: str):
        self.__kafka_topic = value

    @property
    def kafka_connection_string(self) -> str:
        """Kafka connection string with hosts and ports to connect."""
        return self.__kafka_connection_string

    @kafka_connection_string.setter
    def kafka_connection_string(self, value: str):
        input_value = value or environment.get_variable(
            "KAFKA_CONSUMER_CONNECTION_STRING"
        )
        if input_value is None:
            raise ValueError("Config 'kafka connection string' cannot be empty.")
        self.__kafka_connection_string = input_value

    @property
    def mode(self) -> str:
        """Write mode for Spark."""
        return self.__mode

    @mode.setter
    def mode(self, value):
        self.__mode = value or "append"

    @property
    def format_(self) -> str:
        """Write format for Spark."""
        return self.__format

    @format_.setter
    def format_(self, value: str):
        self.__format = value or "kafka"

    @property
    def stream_output_mode(self) -> str:
        """Specify the mode from writing streaming data."""
        return self.__stream_output_mode

    @stream_output_mode.setter
    def stream_output_mode(self, value: str):
        self.__stream_output_mode = value or "update"

    @property
    def stream_checkpoint_path(self) -> str:
        """Path on S3 to save checkpoints for the stream job."""
        return self.__stream_checkpoint_path

    @stream_checkpoint_path.setter
    def stream_checkpoint_path(self, value: str):
        self.__stream_checkpoint_path = value or environment.get_variable(
            "STREAM_CHECKPOINT_PATH"
        )

    @property
    def stream_processing_time(self) -> str:
        """Processing time interval for streaming jobs."""
        return self.__stream_processing_time

    @stream_processing_time.setter
    def stream_processing_time(self, value: str):
        self.__stream_processing_time = value or "0 seconds"

    def get_options(self, topic: str) -> dict:
        """Get options for connecting to Kafka.

        Options will be a dictionary with the write and read configuration for
        spark to kafka.

        Args:
            topic: topic related to Kafka.

        Returns:
            Configuration to connect to Kafka.

        """
        return {
            "topic": self.kafka_topic or topic,
            "kafka.bootstrap.servers": self.kafka_connection_string,
        }

    def translate(self, schema) -> List[Dict]:
        """Get feature set schema to be translated.

        The output will be a list of dictionaries regarding cassandra
        database schema.

        Args:
            schema: feature set schema in spark.

        Returns:
            Kafka schema.

        """
        pass



"""DataType Enum Entity."""

from enum import Enum

from pyspark.sql.types import ArrayType, BinaryType, BooleanType
from pyspark.sql.types import DataType as PySparkDataType
from pyspark.sql.types import (
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    StringType,
    TimestampType,
)
from typing_extensions import final


@final
class DataType(Enum):
    """Holds constants for data types within Butterfree."""

    TIMESTAMP = (TimestampType(), "timestamp")
    BINARY = (BinaryType(), "boolean")
    BOOLEAN = (BooleanType(), "boolean")
    DATE = (DateType(), "timestamp")
    DECIMAL = (DecimalType(), "decimal")
    DOUBLE = (DoubleType(), "double")
    FLOAT = (FloatType(), "float")
    INTEGER = (IntegerType(), "int")
    BIGINT = (LongType(), "bigint")
    STRING = (StringType(), "text")
    ARRAY_BIGINT = (ArrayType(LongType()), "frozen<list<bigint>>")
    ARRAY_STRING = (ArrayType(StringType()), "frozen<list<text>>")
    ARRAY_FLOAT = (ArrayType(FloatType()), "frozen<list<float>>")

    def __init__(self, spark: PySparkDataType, cassandra: str) -> None:
        self.spark = spark
        self.cassandra = cassandra



"""KafkaSource entity."""

from pyspark.sql.dataframe import DataFrame, StructType
from pyspark.sql.functions import col, struct

from butterfree.clients import SparkClient
from butterfree.configs import environment
from butterfree.extract.pre_processing import explode_json_column
from butterfree.extract.readers.reader import Reader


class KafkaReader(Reader):
    """Responsible for get data from a Kafka topic.

    Attributes:
        id: unique string id for register the reader as a view on the metastore
        value_schema: expected schema of the default column named "value" from Kafka.
        topic: string with the Kafka topic name to subscribe.
        connection_string: string with hosts and ports to connect.
            The string need to be in the format: host1:port,host2:port,...,hostN:portN.
            The argument is not necessary if is passed as a environment variable
            named KAFKA_CONSUMER_CONNECTION_STRING.
        topic_options: additional options for consuming from topic. See docs:
            https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html.
        stream: flag to indicate the reading mode: stream or batch

    The default df schema coming from Kafka reader of Spark is the following:
        key:string
        value:string
        topic:string
        partition:integer
        offset:long
        timestamp:timestamp
        timestampType:integer

    But using this reader and passing the desired schema under value_schema we would
    have the following result:

    With value_schema declared as:

    >>> value_schema = StructType(
    ...     [
    ...         StructField("ts", LongType(), nullable=True),
    ...         StructField("id", LongType(), nullable=True),
    ...         StructField("type", StringType(), nullable=True),
    ...     ]
    ... )

    The output df schema would be:
        ts:long
        id:long
        type:string
        kafka_metadata:struct
            key:string
            topic:string
            value:string
            partition:integer
            offset:long
            timestamp:timestamp
            timestampType:integer

    Instantiation example:

    >>> from butterfree.extract.readers import KafkaReader
    >>> from butterfree.clients import SparkClient
    >>> from pyspark.sql.types import StructType, StructField, StringType, LongType
    >>> spark_client = SparkClient()
    >>> value_schema = StructType(
    ...     [
    ...         StructField("ts", LongType(), nullable=True),
    ...         StructField("id", LongType(), nullable=True),
    ...         StructField("type", StringType(), nullable=True),
    ...     ]
    ... )
    >>> kafka_reader = KafkaReader(
    ...                 id="kafka_reader_id",
    ...                 topic="topic",
    ...                 value_schema=value_schema
    ...                 connection_string="host1:port,host2:port",
    ...                )
    >>> df = kafka_reader.consume(spark_client)

    This last method will use the Spark Client, as default, to read
    the desired topic, loading data into a dataframe, according to
    KafkaReader class arguments.

    In this case, however, a temp view will be created, containing
    the transformed data.

    """

    KAFKA_COLUMNS = [
        "key",
        "topic",
        "value",
        "partition",
        "offset",
        "timestamp",
        "timestampType",
    ]

    __name__ = "Kafka Reader"

    def __init__(
        self,
        id: str,
        topic: str,
        value_schema: StructType,
        connection_string: str = None,
        topic_options: dict = None,
        stream: bool = True,
    ):
        super().__init__(id)
        if not isinstance(topic, str):
            raise ValueError("topic must be a string with the topic name")
        if not isinstance(value_schema, StructType):
            raise ValueError(
                "value_schema must be a StructType with the schema "
                'of the JSON presented in "value" Kafka column'
            )
        self.topic = topic
        self.value_schema = value_schema
        self.connection_string = connection_string or environment.get_variable(
            "KAFKA_CONSUMER_CONNECTION_STRING"
        )
        self.options = dict(
            {
                "kafka.bootstrap.servers": self.connection_string,
                "subscribe": self.topic,
            },
            **topic_options if topic_options else {},
        )
        self.stream = stream

    def _struct_df(self, df: DataFrame) -> DataFrame:
        """Struct the output dataframe generated by the reader.

        Under the default "value" column coming from Kafka there are the custom
        fields created by some producer. This function will struct the dataframe as
        to get all desired fields from "value" and insert all Kafka default columns,
        including "value", under "kafka_metadata" column. It is important to notice
        that the declared value_schema suffer from the same effects described in
        explode_json_column method in pre_processing module.

        Args:
            df: direct dataframe output from from KafkaReader.

        Returns:
            Structured dataframe with kafka value fields as columns.
                All other default fields from Kafka will be stored under
                "kafka_metadata" column.

        """
        df = df.withColumn("kafka_metadata", struct(*self.KAFKA_COLUMNS))
        df = explode_json_column(df, column="value", json_schema=self.value_schema)
        return df.select(
            [field.name for field in self.value_schema] + ["kafka_metadata"]
        )

    def consume(self, client: SparkClient) -> DataFrame:
        """Extract data from a kafka topic.

        When stream mode it will get all the new data arriving at the topic in a
        streaming dataframe. When not in stream mode it will get all data
        available in the kafka topic.

        Args:
            client: client responsible for connecting to Spark session.

        Returns:
            Dataframe with data from topic.

        """
        # read using client and cast key and value columns from binary to string
        raw_df = (
            client.read(format="kafka", options=self.options, stream=self.stream)
            .withColumn("key", col("key").cast("string"))
            .withColumn("value", col("value").cast("string"))
        )

        # apply schema defined in self.value_schema
        return self._struct_df(raw_df)



"""FeatureSetPipeline entity."""
from typing import List

from butterfree.clients import SparkClient
from butterfree.dataframe_service import repartition_sort_df
from butterfree.extract import Source
from butterfree.load import Sink
from butterfree.transform import FeatureSet


class FeatureSetPipeline:
    """Defines a ETL pipeline for the construction of a feature set.

    Attributes:
        source: source of the data, the entry point of the pipeline.
        feature_set: feature set composed by features and context metadata.
        sink: sink used to write the output dataframe in the desired locations.
        spark_client: client used to access Spark connection.

    Example:
        This an example regarding the feature set pipeline definition. All
        sources, feature set (and its features) and writers are defined.

        >>> import os

        >>> from butterfree.pipelines import FeatureSetPipeline
        >>> from butterfree.constants.columns import TIMESTAMP_COLUMN
        >>> from butterfree.configs.db import MetastoreConfig
        >>> from butterfree.extract import Source
        >>> from butterfree.extract.readers import TableReader
        >>> from butterfree.transform import FeatureSet
        >>> from butterfree.transform.features import (
        ...     Feature,
        ...     KeyFeature,
        ...     TimestampFeature,
        ...)
        >>> from butterfree.transform.transformations import (
        ...     SparkFunctionTransform,
        ...     CustomTransform,
        ... )
        >>> from butterfree.load import Sink
        >>> from butterfree.load.writers import HistoricalFeatureStoreWriter
        >>> import pyspark.sql.functions as F

        >>> def divide(df, fs, column1, column2):
        ...     name = fs.get_output_columns()[0]
        ...     df = df.withColumn(name, F.col(column1) / F.col(column2))
        ...     return df

        >>> pipeline = FeatureSetPipeline(
        ...    source=Source(
        ...        readers=[
        ...            TableReader(
        ...                id="table_reader_id",
        ...                database="table_reader_db",
        ...                table="table_reader_table",
        ...            ),
        ...        ],
        ...        query=f"select * from table_reader_id ",
        ...    ),
        ...    feature_set=FeatureSet(
        ...        name="feature_set",
        ...        entity="entity",
        ...        description="description",
        ...        features=[
        ...            Feature(
        ...                name="feature1",
        ...                description="test",
        ...            transformation=SparkFunctionTransform(
        ...                 functions=[F.avg, F.stddev_pop]
        ...             ).with_window(
        ...                 partition_by="id",
        ...                 order_by=TIMESTAMP_COLUMN,
        ...                 mode="fixed_windows",
        ...                 window_definition=["2 minutes", "15 minutes"],
        ...             ),
        ...            ),
        ...            Feature(
        ...                name="divided_feature",
        ...                description="unit test",
        ...                transformation=CustomTransform(
        ...                    transformer=divide,
        ...                    column1="feature1",
        ...                    column2="feature2",
        ...                ),
        ...            ),
        ...        ],
        ...        keys=[
        ...            KeyFeature(
        ...                name="id",
        ...                description="The user's Main ID or device ID"
        ...            )
        ...        ],
        ...        timestamp=TimestampFeature(),
        ...    ),
        ...    sink=Sink(
        ...         writers=[
        ...            HistoricalFeatureStoreWriter(
        ...                db_config=MetastoreConfig(
        ...                    format_="parquet",
        ...                    path=os.path.join(
        ...                        os.path.dirname(os.path.abspath(__file__))
        ...                    ),
        ...                ),
        ...            )
        ...        ],
        ...    ),
        ...)

        >>> pipeline.run()

        This last method (run) will execute the pipeline flow, it'll read from
        the defined sources, compute all the transformations and save the data
        to the specified locations.

    """

    def __init__(
        self,
        source: Source,
        feature_set: FeatureSet,
        sink: Sink,
        spark_client: SparkClient = None,
    ):
        self.source = source
        self.feature_set = feature_set
        self.sink = sink
        self.spark_client = spark_client

    @property
    def source(self) -> Source:
        """Source of the data, the entry point of the pipeline."""
        return self._source

    @source.setter
    def source(self, source: Source):
        if not isinstance(source, Source):
            raise ValueError("source must be a Source instance")
        self._source = source

    @property
    def feature_set(self) -> FeatureSet:
        """Feature set composed by features and context metadata."""
        return self._feature_set

    @feature_set.setter
    def feature_set(self, feature_set: FeatureSet):
        if not isinstance(feature_set, FeatureSet):
            raise ValueError("feature_set must be a FeatureSet instance")
        self._feature_set = feature_set

    @property
    def sink(self):
        """Sink used to write the output dataframe in the desired locations."""
        return self._sink

    @sink.setter
    def sink(self, sink: Sink):
        if not isinstance(sink, Sink):
            raise ValueError("sink must be a Sink instance")
        self._sink = sink

    @property
    def spark_client(self):
        """Client used to access Spark connection."""
        return self._spark_client

    @spark_client.setter
    def spark_client(self, spark_client: SparkClient):
        self._spark_client = spark_client or SparkClient()
        if not isinstance(self._spark_client, SparkClient):
            raise ValueError("spark_client must be a SparkClient instance")

    def run(
        self,
        end_date: str = None,
        partition_by: List[str] = None,
        order_by: List[str] = None,
        num_processors: int = None,
    ):
        """Runs the defined feature set pipeline.

        The pipeline consists in the following steps:
        - Constructs the input dataframe from the data source.
        - Construct the feature set dataframe using the defined Features.
        - Load the data to the configured sink locations.

        It's important to notice, however, that both parameters partition_by
        and num_processors are WIP, we intend to enhance their functionality
        soon. Use only if strictly necessary.

        """
        dataframe = self.source.construct(client=self.spark_client)

        if partition_by:
            order_by = order_by or partition_by
            dataframe = repartition_sort_df(
                dataframe, partition_by, order_by, num_processors
            )

        dataframe = self.feature_set.construct(
            dataframe=dataframe,
            client=self.spark_client,
            end_date=end_date,
            num_processors=num_processors,
        )

        self.sink.flush(
            dataframe=dataframe,
            feature_set=self.feature_set,
            spark_client=self.spark_client,
        )

        if not dataframe.isStreaming:
            self.sink.validate(
                dataframe=dataframe,
                feature_set=self.feature_set,
                spark_client=self.spark_client,
            )



"""Module where there are repartition methods."""
from typing import List

from pyspark.sql.dataframe import DataFrame

from butterfree.constants.spark_constants import (
    DEFAULT_NUM_PARTITIONS,
    PARTITION_PROCESSOR_RATIO,
)


def _num_partitions_definition(num_processors, num_partitions):
    num_partitions = (
        num_processors * PARTITION_PROCESSOR_RATIO
        if num_processors
        else num_partitions or DEFAULT_NUM_PARTITIONS
    )

    return num_partitions


def repartition_df(
    dataframe: DataFrame,
    partition_by: List[str],
    num_partitions: int = None,
    num_processors: int = None,
):
    """Partition the DataFrame.

    Args:
        dataframe: Spark DataFrame.
        partition_by: list of partitions.
        num_processors: number of processors.
        num_partitions: number of partitions.

    Returns:
        Partitioned dataframe.

    """
    num_partitions = _num_partitions_definition(num_processors, num_partitions)
    return dataframe.repartition(num_partitions, *partition_by)


def repartition_sort_df(
    dataframe: DataFrame,
    partition_by: List[str],
    order_by: List[str],
    num_processors: int = None,
    num_partitions: int = None,
):
    """Partition and Sort the DataFrame.

    Args:
        dataframe: Spark DataFrame.
        partition_by: list of columns to partition by.
        order_by: list of columns to order by.
        num_processors: number of processors.
        num_partitions: number of partitions.

    Returns:
        Partitioned and sorted dataframe.

    """
    num_partitions = _num_partitions_definition(num_processors, num_partitions)
    dataframe = repartition_df(dataframe, partition_by, num_partitions)
    return dataframe.sortWithinPartitions(*order_by)



"""Method to compute mode aggregation."""
import pandas as pd
from pyspark.sql.functions import pandas_udf
from pyspark.sql.types import StringType


@pandas_udf(StringType())
def mode(column: pd.Series) -> str:
    """Computes a mode aggregation.

    Attributes:
        column: desired data to be aggregated with mode.

    Example:
        It's necessary to declare the desired aggregation method, (average,
        standard deviation and count are currently supported, as it can be
        seen in __ALLOWED_AGGREGATIONS) and, finally, define the mode.

        >>> from pyspark import SparkContext
        >>> from pyspark.sql import session, Window
        >>> from pyspark.sql.functions import pandas_udf
        >>> from butterfree.transform\
        ...      .transformations.user_defined_functions import (mode)
        >>> sc = SparkContext.getOrCreate()
        >>> spark = session.SparkSession(sc)
        >>> df = spark.createDataFrame(
        >>>      [(1, 1), (1, 1), (2, 2), (2, 1), (2, 2)],
        >>>      ("id", "column"))
        >>> df.groupby("id").agg(mode("column")).show()
        +---+------------+
        | id|mode(column)|
        +---+------------+
        |  1|           1|
        |  2|           2|
        +---+------------+
        >>> w = Window.partitionBy('id').rowsBetween(
        ...       Window.unboundedPreceding, Window.unboundedFollowing)
        >>> df.withColumn('most_viewed', mode("column").over(w)).show()
        +---+------+-----------+
        | id|column|most_viewed|
        +---+------+-----------+
        |  1|     1|          1|
        |  1|     1|          1|
        |  2|     2|          2|
        |  2|     1|          2|
        |  2|     2|          2|
        +---+------+-----------+

        This example shows the mode aggregation. It's important to notice,
        however, that if we want to used in fixed_windows or row_windows mode,
        we'd need unbounded windows. For that reason, mode is meant to be used
        just in rolling_windows mode, initially. We intend to make it available
        to others modes soon.

    """
    return str(column.mode()[0])



"""Dataframe optimization components regarding Butterfree."""
from butterfree.dataframe_service.repartition import repartition_df, repartition_sort_df

__all__ = ["repartition_df", "repartition_sort_df"]



"""SparkClient entity."""

from typing import Any, Dict, List, Optional, Union

from pyspark.sql import DataFrame, DataFrameReader, SparkSession
from pyspark.sql.streaming import DataStreamReader, StreamingQuery
from pyspark.sql.types import StructType

from butterfree.clients import AbstractClient


class SparkClient(AbstractClient):
    """Handle Spark session connection.

    Get query results with SQL, reads and writes data on external systems.

    """

    def __init__(self) -> None:
        self._session: Optional[SparkSession] = None

    @property
    def conn(self) -> SparkSession:
        """Gets or creates an SparkSession.

        Returns:
            Spark session

        """
        if not self._session:
            self._session = SparkSession.builder.getOrCreate()
        return self._session

    def read(
        self,
        format: str,
        options: Dict[str, Any],
        schema: Optional[StructType] = None,
        stream: bool = False,
    ) -> DataFrame:
        """Use the SparkSession.read interface to load data into a dataframe.

        Check docs for more information:
            https://spark.apache.org/docs/latest/sql-data-sources-load-save-functions.html#generic-loadsave-functions

        Args:
            format: string with the format to be used by the DataframeReader.
            options: options to setup the DataframeReader.
            stream:  flag to indicate if data must be read in stream mode.
            schema: an optional pyspark.sql.types.StructType for the input schema.

        Returns:
            Dataframe

        """
        if not isinstance(format, str):
            raise ValueError("format needs to be a string with the desired read format")
        if not isinstance(options, dict):
            raise ValueError("options needs to be a dict with the setup configurations")

        df_reader: Union[
            DataStreamReader, DataFrameReader
        ] = self.conn.readStream if stream else self.conn.read
        df_reader = df_reader.schema(schema) if schema else df_reader
        return df_reader.format(format).options(**options).load()

    def read_table(self, table: str, database: str = None) -> DataFrame:
        """Use the SparkSession.read interface to read a metastore table.

        Args:
            database: name of the metastore database/schema
            table: name of the table in metastore

        Returns:
            Dataframe

        """
        if not isinstance(table, str):
            raise ValueError(
                "table needs to be a string with the name of the registered table"
            )
        return self.conn.read.table(f"{database}.{table}" if database else table)

    def sql(self, query: str) -> DataFrame:
        """Run a query using Spark SQL.

        Args:
            query: Spark SQL query.

        Returns:
            Dataframe

        """
        return self.conn.sql(query)

    @staticmethod
    def write_dataframe(
        dataframe: DataFrame, format_: str, mode: str, **options: Any
    ) -> None:
        """Receive a spark DataFrame and write it.

        Args:
            dataframe: dataframe containing data from a feature set.
            format_: format used to save the dataframe.
            mode: writing modem can be "error", "append", "overwrite" or
                "ignore". For more information:
                [here](https://spark.apache.org/docs/2.3.0/sql-programming-guide.html#save-modes).
            **options: all other options that can be used in a DataFrameWriter.

        """
        if not isinstance(format_, str):
            raise ValueError("format needs to be a string")
        if not isinstance(mode, str):
            raise ValueError("mode needs to be a string")
        dataframe.write.save(format=format_, mode=mode, **options)

    def write_stream(
        self,
        dataframe: DataFrame,
        processing_time: str,
        output_mode: str,
        checkpoint_path: str,
        format_: str,
        mode: str,
        **options: Any,
    ) -> StreamingQuery:
        """Starts streaming data writing job.

        Args:
            dataframe: Spark dataframe containing data from a feature set.
            processing_time: a processing time interval as a string.
                E.g. '5 seconds', '1 minute'. Set a trigger that runs the
                mini-batch periodically based on the processing time. If the
                effect of processing data as soon as the data arrives, without
                having to wait for the time frame, is desired, the value
                '0 seconds' can be set.
            output_mode: specifies how data of a streaming DataFrame/Dataset is
                written to a streaming sink destination.
            checkpoint_path: path on S3 to save checkpoints for the stream job.
                These checkpoint can be used on the the job re-start to return
                from where it stops.
            format_: format used to save the dataframe.
            mode: writing modem can be "error", "append", "overwrite" or
                "ignore". For more information:
                [here](https://spark.apache.org/docs/2.3.0/sql-programming-guide.html#save-modes).
            **options: all other options that can be used in a DataFrameWriter.

        More information about processing_time, output_mode and checkpoint_path
        can be found in Spark documentation:
        [here](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html)

        Returns:
            Streaming handler.

        """
        if not dataframe.isStreaming:
            raise ValueError("A stream df is needed to start a streaming job.")

        return (
            dataframe.writeStream.trigger(processingTime=processing_time)
            .outputMode(output_mode)
            .option("checkpointLocation", checkpoint_path)
            .foreachBatch(
                lambda batch_df, _: self.write_dataframe(
                    batch_df, format_, mode, **options
                )
            )
            .start()
        )

    @staticmethod
    def write_table(
        dataframe: DataFrame,
        database: str,
        table_name: str,
        path: str,
        format_: str = None,
        mode: str = None,
        partition_by: List[str] = None,
        **options: Any,
    ) -> None:
        """Receive a spark DataFrame and write it as a table in metastore.

        Args:
            dataframe: spark dataframe containing data from a feature set.
            database: specified database name.
            table_name: specified table name.
            path: string with the local to save the table.
            format_: string with the format used to save.
            mode: writing mode, it can be: "error", "append", "overwrite" or
                "ignore". More information:
                [here](https://spark.apache.org/docs/2.3.0/sql-programming-guide.html#save-modes).
            partition_by: names of partitioning columns.
            options: all other options that can be used in a DataFrameWriter.

        """
        if not isinstance(database, str):
            raise ValueError("database needs to be a string")
        if not isinstance(table_name, str):
            raise ValueError("table_name needs to be a string")
        if not isinstance(path, str):
            raise ValueError("path needs to be a string of the local to save")

        name = "{}.{}".format(database, table_name)

        dataframe.write.saveAsTable(
            mode=mode,
            format=format_,
            partitionBy=partition_by,
            name=name,
            path=path,
            **options,
        )

    def create_temporary_view(
        self, dataframe: DataFrame, name: str
    ) -> Optional[StreamingQuery]:
        """Create a temporary view from a given dataframe.

        Args:
            dataframe: dataframe to be be queried by the view.
            name: name of the temporary view.

        """
        if not dataframe.isStreaming:
            return dataframe.createOrReplaceTempView(name)
        return dataframe.writeStream.format("memory").queryName(name).start()



"""Holds common spark constants, present through all Butterfree."""

from typing_extensions import Final

# from spark.sql.shuffle.partitions default value
DEFAULT_NUM_PARTITIONS: Final = 200

# ratio between number of partitions per processor recommended (lower bound: 2)
# refs:
# https://github.com/vaquarkhan/Apache-Kafka-poc-and-notes/wiki/Apache-Spark-Join-guidelines-and-Performance-tuning
PARTITION_PROCESSOR_RATIO: Final = 4



"""Utils for date range generation."""

from datetime import datetime
from typing import Union

from pyspark.sql import functions

from butterfree.clients import SparkClient
from butterfree.constants import DataType
from butterfree.constants.columns import TIMESTAMP_COLUMN


def get_date_range(
    client: SparkClient,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    step: int = None,
):
    """Create a date range dataframe.

    The dataframe returning from this method will containing a single column
    TIMESTAMP_COLUMN, of timestamp type, with dates between start and end.

    Args:
        client: a spark client.
        start_date: range beginning value (inclusive).
        end_date: range last value (exclusive)
        step: optional step, in seconds.

    Returns:
        A single column date range spark dataframe.
    """
    day_in_seconds = 60 * 60 * 24
    step = step or day_in_seconds
    start_date = (
        start_date if isinstance(start_date, str) else start_date.strftime("%Y-%m-%d")
    )
    end_date = end_date if isinstance(end_date, str) else end_date.strftime("%Y-%m-%d")
    date_df = client.conn.createDataFrame(
        [(start_date, end_date)], ("start_date", "end_date")
    ).select(
        [
            functions.col(c).cast(DataType.TIMESTAMP.spark).cast(DataType.BIGINT.spark)
            for c in ("start_date", "end_date")
        ]
    )
    start_date, end_date = date_df.first()
    return client.conn.range(start_date, end_date + day_in_seconds, step).select(
        functions.col("id").cast(DataType.TIMESTAMP.spark).alias(TIMESTAMP_COLUMN)
    )



"""Feature entity."""
import warnings
from typing import List

from parameters_validation import non_blank, non_null
from pyspark.sql import DataFrame
from pyspark.sql.functions import col

from butterfree.constants import DataType
from butterfree.transform.transformations import (
    AggregatedTransform,
    CustomTransform,
    SparkFunctionTransform,
    SQLExpressionTransform,
    StackTransform,
    TransformComponent,
)
from butterfree.transform.transformations.h3_transform import H3HashTransform


class Feature:
    """Defines a Feature.

    A Feature is the result of a transformation over one (or more) data columns
    over an input dataframe. Transformations can be as simple as renaming,
    casting types, mathematical expressions or complex functions/models.

    Attributes:
        name: feature name.
            Can be use by the transformation to derive multiple output columns.
        description: brief explanation regarding the feature.
        dtype: data type for the output columns of this feature.
        from_column: original column to build feature.
            Used when there is transformation or the transformation has no
            reference about the column to use for.
        transformation: transformation that will be applied to create this
            feature.

    """

    def __init__(
        self,
        name: non_blank(str),
        description: non_blank(str),
        dtype: non_blank(DataType) = None,
        from_column: non_blank(str) = None,
        transformation: non_null(TransformComponent) = None,
    ) -> None:
        self.name = name
        self.description = description
        self.transformation = transformation
        self.dtype = dtype
        self.from_column = from_column

    @property
    def dtype(self) -> DataType:
        """Attribute dtype getter."""
        return self.__dtype

    @dtype.setter
    def dtype(self, value: DataType) -> None:
        if (
            not (
                isinstance(self.transformation, AggregatedTransform)
                or isinstance(self.transformation, SparkFunctionTransform)
            )
            and value is None
        ):
            raise ValueError(
                "dtype can't be None, except if the transformation is "
                "AggregatedTransform or SparkFunctionTransform"
            )

        if value is not None and not isinstance(value, DataType):
            raise ValueError("dtype must be a DataType.")

        self.__dtype = value

    @property
    def from_column(self) -> str:
        """Attribute from_column getter."""
        return self.__from_column

    @from_column.setter
    def from_column(self, value: str) -> None:
        if (
            isinstance(self.transformation, CustomTransform)
            or isinstance(self.transformation, SQLExpressionTransform)
            or isinstance(self.transformation, H3HashTransform)
            or isinstance(self.transformation, StackTransform)
        ) and value is not None:
            raise ValueError(
                f"from_column parameter doesn't need to be declared, "
                f"the column name {value} would be overwritten with "
                f"another name."
            )

        if value is not None and not isinstance(value, str):
            raise ValueError("from_column must be a string.")

        self.__from_column = value

    @property
    def transformation(self) -> TransformComponent:
        """Attribute transformation getter."""
        return self.__transformation

    @transformation.setter
    def transformation(self, value: TransformComponent) -> None:
        self.__transformation = value

        if value is not None:
            self.__transformation.parent = self

    def get_output_columns(self) -> List[str]:
        """Get output columns that will be generated by this feature engineering.

        Returns
            Output columns names.
        """
        if self.transformation is not None:
            return self.transformation.output_columns
        return [self.name]

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe for the transformation.

        Returns:
            Transformed dataframe.
        """
        if self.transformation:
            return self.transformation.transform(dataframe)

        if self.from_column:
            if self.name in dataframe.columns:
                warnings.warn(
                    f"The column name {self.name} "
                    "already exists in the dataframe and "
                    "will be overwritten with another column."
                )

            dataframe = dataframe.withColumn(self.name, col(self.from_column))

        if self.dtype:
            dataframe = dataframe.withColumn(
                self.name, col(self.name).cast(self.dtype.spark)
            )
        return dataframe



"""H3 Transform entity."""

from typing import List

from parameters_validation import non_blank
from pyspark.sql import DataFrame
from pyspark.sql.functions import lit, udf
from pyspark.sql.types import StringType

from butterfree.transform.transformations.stack_transform import StackTransform
from butterfree.transform.transformations.transform_component import TransformComponent

try:
    from h3 import h3
except ModuleNotFoundError as e:
    e.msg = (
        "H3 not found. To be able to use this module,"
        "you must install butterfree[h3]."
    )
    raise


@udf(StringType())
def define_h3(lat, lng, resolution):
    """UDF for h3 hash retrieval.

    Attributes:
        lat: latitude column.
        lng: longitude column.
        resolution: desired h3 resolution.
    """
    if lat and lng:
        h3_feature = h3.geo_to_h3(lat, lng, resolution)
    else:
        h3_feature = None
    return h3_feature


class H3HashTransform(TransformComponent):
    """Defines a H3 hash transformation.

    Attributes:
        h3_resolutions: h3 resolutions from 6 to 12.
        lat_column: latitude column.
        lng_column: longitude column.

    Example:
        It's necessary to declare the desired h3 resolutions and
        latitude and longitude columns.

        >>> from butterfree.transform.features import Feature
        >>> from butterfree.transform.transformations.h3_transform import (
        ... H3HashTransform
        ...)
        >>> from butterfree.constants import DataType
        >>> from pyspark import SparkContext
        >>> from pyspark.sql import session
        >>> sc = SparkContext.getOrCreate()
        >>> spark = session.SparkSession(sc)
        >>> df = spark.createDataFrame([(1, 200, -23.554190, -46.670723),
        ...                             (1, 300, -23.554190, -46.670723),
        ...                             (1, 400, -23.554190, -46.670723),
        ...                             (1, 500, -23.554190, -46.670723)]
        ...                           ).toDF("id", "feature", "lat", "lng")
        >>> feature = Feature(
        ...    name="feature",
        ...    description="h3 hash transform usage example",
        ...    dtype=DataType.STRING,
        ...    transformation=H3HashTransform(
        ...        h3_resolutions=[6, 7, 8, 9, 10, 11, 12],
        ...        lat_column="lat",
        ...        lng_column="lng",
        ...    )
        ... )
        >>> feature.transform(df).show()
        +-------+---+---------+----------+-------------------+-------------------+
        |feature| id|      lat|       lng|lat_lng__h3_hash__6|lat_lng__h3_hash__7|
        +-------+---+---------+----------+-------------------+-------------------+
        |    200|  1|-23.55419|-46.670723|    86a8100efffffff|    87a8100eaffffff|
        |    300|  1|-23.55419|-46.670723|    86a8100efffffff|    87a8100eaffffff|
        |    400|  1|-23.55419|-46.670723|    86a8100efffffff|    87a8100eaffffff|
        |    500|  1|-23.55419|-46.670723|    86a8100efffffff|    87a8100eaffffff|
        +-------+---+---------+----------+-------------------+-------------------+

    """

    def __init__(
        self,
        h3_resolutions: non_blank(List[int]),
        lat_column: non_blank(str),
        lng_column: non_blank(str),
    ):
        super().__init__()
        self.h3_resolutions = h3_resolutions
        self.lat_column = lat_column
        self.lng_column = lng_column
        self.stack_transform = None

    @property
    def output_columns(self) -> List[str]:
        """Columns generated by the transformation."""
        output_columns = []
        for h3_resolution in self.h3_resolutions:
            output_columns.append(f"lat_lng__h3_hash__{h3_resolution}")

        if self.stack_transform:
            return [self._parent.name]

        return output_columns

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe.

        Returns:
            Transformed dataframe.

        """
        for h3_resolution in self.h3_resolutions:
            dataframe = dataframe.withColumn(
                f"lat_lng__h3_hash__{h3_resolution}",
                define_h3(self.lat_column, self.lng_column, lit(h3_resolution)),
            )
        if self.stack_transform:
            self.stack_transform._parent = self._parent
            return self.stack_transform.transform(dataframe)
        return dataframe

    def with_stack(self):
        """Add a final Stack step to the transformation.

        A new column will be created stacking all the the resolution columns
        generated by H3. The name of this column will be the parent name of
        H3Transform, for example, the name of the KeyFeature that has the H3
        as a transformation.

        """
        self.stack_transform = StackTransform(*self.output_columns)
        return self
